from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Generator, Generic, NoReturn, Optional

from typing_extensions import ParamSpec

from sqs_workers.utils import bind_arguments

if TYPE_CHECKING:
    from sqs_workers.queue import JobQueue


P = ParamSpec("P")


class AsyncTask(Generic[P]):
    def __init__(
        self, queue: JobQueue, job_name: str, processor: Callable[P, Any]
    ) -> None:
        self.queue = queue
        self.job_name = job_name
        self.processor = processor
        self.__doc__ = processor.__doc__

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> NoReturn:
        raise RuntimeError(
            f"Async task {self.queue.name}.{self.job_name} called synchronously "
            "(probably, by mistake). Use either AsyncTask.run(...) "
            "to run the task synchronously or AsyncTask.delay(...) "
            "to add it to the queue"
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.queue.name}.{self.job_name}>"

    def run(self, *args: P.args, **kwargs: P.kwargs) -> Any:
        """Run the task synchronously."""
        if self.queue.batching_policy.batching_enabled:
            if len(args) > 0:
                raise TypeError("Must use keyword arguments only for batch read queues")
            kwargs = bind_arguments(self.processor, [[kwargs]], {})
            return self.processor(**kwargs)
        else:
            kwargs = bind_arguments(self.processor, args, kwargs)
            return self.processor(**kwargs)

    @contextmanager
    def batch(self) -> Generator[None, None, None]:
        """Context manager to add jobs in batch."""
        with self.queue.add_batch():
            yield

    def delay(self, *args: P.args, **kwargs: P.kwargs) -> Optional[str]:
        """Run the task asynchronously."""
        _content_type = kwargs.pop("_content_type", self.queue.env.codec)  # type: ignore
        _delay_seconds = kwargs.pop("_delay_seconds", None)  # type: ignore
        _deduplication_id = kwargs.pop("_deduplication_id", None)  # type: ignore
        _group_id = kwargs.pop("_group_id", None)  # type: ignore

        if self.queue.batching_policy.batching_enabled:
            if len(args) > 0:
                raise TypeError("Must use keyword arguments only for batch read queues")
        else:
            kwargs = bind_arguments(self.processor, args, kwargs)

        return self.queue.add_job(
            self.job_name,
            _content_type=_content_type,  # type: ignore
            _delay_seconds=_delay_seconds,  # type: ignore
            _deduplication_id=_deduplication_id,
            _group_id=_group_id,  # type: ignore
            **kwargs,
        )

    def bake(self, *args: P.args, **kwargs: P.kwargs) -> BakedAsyncTask:
        """
        Create a baked version of the async task, which contains the reference
        to a queue and task, as well as all arguments which needs to be passed
        to it.
        """
        return BakedAsyncTask(self, args, kwargs)


class BakedAsyncTask:
    def __init__(self, async_task, args, kwargs):
        self.async_task = async_task
        self.args = args
        self.kwargs = kwargs

    def __call__(self) -> None:
        self.async_task(*self.args, **self.kwargs)

    def delay(self) -> None:
        self.async_task.delay(*self.args, **self.kwargs)

    def __repr__(self) -> str:
        return "BakedAsyncTask(%r, ...)" % self.async_task



"""
In memory mockup implementation of essential parts of boto3 SQS objects.

Used for faster and more predictable processing in tests.

Lacks some features of "real sqs", and some other features implemented
ineffectively.

- Redrive policy doesn't work
- There is no differences between standard and FIFO queues
- FIFO queues don't support content-based deduplication
- Delayed tasks executed ineffectively: the task is gotten from the queue,
  and if the time hasn't come yet, the task is put back.
- API can return slightly different results

Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html
"""
import datetime
import logging
import uuid
from typing import Any, Dict, List

import attr

logger = logging.getLogger(__name__)


@attr.s
class MemoryAWS:
    """In-memory AWS as a whole."""

    client: "MemoryClient" = attr.ib(
        repr=False,
        default=attr.Factory(lambda self: MemoryClient(self), takes_self=True),
    )
    resource: "ServiceResource" = attr.ib(
        repr=False,
        default=attr.Factory(lambda self: ServiceResource(self), takes_self=True),
    )
    queues: List["MemoryQueue"] = attr.ib(factory=list)

    def create_queue(self, QueueName: str, Attributes) -> "MemoryQueue":
        queue = MemoryQueue(self, QueueName, Attributes)
        self.queues.append(queue)
        return queue

    def delete_queue(self, QueueUrl: str) -> None:
        self.queues = [queue for queue in self.queues if queue.url != QueueUrl]


@attr.s
class MemorySession:
    """In memory AWS session."""

    aws = attr.ib(repr=False, factory=MemoryAWS)

    def client(self, service_name: str, **kwargs):
        assert service_name == "sqs"
        return self.aws.client

    def resource(self, service_name: str, **kwargs):
        assert service_name == "sqs"
        return self.aws.resource


@attr.s
class MemoryClient:
    aws = attr.ib(repr=False)

    def create_queue(self, QueueName: str, Attributes):
        return self.aws.create_queue(QueueName, Attributes)

    def delete_queue(self, QueueUrl: str):
        return self.aws.delete_queue(QueueUrl)

    def list_queues(self, QueueNamePrefix=""):
        return {
            "QueueUrls": [
                queue.url
                for queue in self.aws.queues
                if queue.name.startswith(QueueNamePrefix)
            ]
        }


@attr.s
class ServiceResource:
    aws: MemoryAWS = attr.ib(repr=False)

    def create_queue(self, QueueName: str, Attributes):
        return self.aws.create_queue(QueueName, Attributes)

    def get_queue_by_name(self, QueueName: str):
        for queue in self.aws.queues:
            if queue.name == QueueName:
                return queue
        return None


@attr.s
class MemoryQueue:
    """
    In-memory queue which mimics the subset of SQS Queue object.

    Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
         services/sqs.html#queue
    """

    aws: MemoryAWS = attr.ib(repr=False)
    name: str = attr.ib()
    attributes: Dict[str, Dict[str, str]] = attr.ib()
    messages: List["MemoryMessage"] = attr.ib(factory=list)
    in_flight: Dict[str, "MemoryMessage"] = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        self.attributes["QueueArn"] = self.name

    @property
    def url(self):
        return f"memory://{self.name}"

    def send_message(self, **kwargs):
        message = MemoryMessage.from_kwargs(self, kwargs)
        self.messages.append(message)
        return {"MessageId": message.message_id, "SequenceNumber": 0}

    def send_messages(self, Entries):
        res = []
        for message in Entries:
            res.append(self.send_message(**message))
        return {"Successful": res, "Failed": []}

    def receive_messages(self, WaitTimeSeconds="0", MaxNumberOfMessages="10", **kwargs):
        """
        Helper function which returns at most max_messages from the
        queue. Used in an infinite loop inside `get_raw_messages`
        """
        wait_seconds = int(WaitTimeSeconds)
        max_messages = int(MaxNumberOfMessages)

        ready_messages = []
        push_back_messages = []

        now = datetime.datetime.utcnow()
        threshold = now + datetime.timedelta(seconds=wait_seconds)

        # before retrieving messages, go through in_flight and return any
        # messages whose "invisible" timeout has expired back to the pool
        for message_id, message in self.in_flight.copy().items():
            if message.visible_at <= threshold:
                self.in_flight.pop(message_id)
                self.messages.append(message)

        for message in self.messages:
            if message.visible_at > threshold or len(ready_messages) >= max_messages:
                push_back_messages.append(message)
            else:
                ready_messages.append(message)

        self.messages[:] = push_back_messages

        # now, mark all returned messages as "in flight" and unavailable
        # to be returned until their invisibility timeout comes (1s)
        for m in ready_messages:
            self.in_flight[m.message_id] = m
            m.change_visibility(VisibilityTimeout="1")

        return ready_messages

    def delete_messages(self, Entries):
        """
        Delete messages implementation.

        See: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
             services/sqs.html#SQS.Queue.delete_messages
        """
        found_entries = []
        not_found_entries = []

        for e in Entries:
            if e["Id"] in self.in_flight:
                found_entries.append(e)
                self.in_flight.pop(e["Id"])
            else:
                not_found_entries.append(e)

        return {
            "Successful": [{"Id": e["Id"]} for e in found_entries],
            "Failed": [{"Id": e["Id"]} for e in not_found_entries],
        }

    def change_message_visibility_batch(self, Entries):
        """
        Changes message visibility by looking at in-flight messages, setting
        a new visible_at, when it will return to the pool of messages
        """
        found_entries = []
        not_found_entries = []

        now = datetime.datetime.utcnow()

        for e in Entries:
            if e["Id"] in self.in_flight:
                found_entries.append(e)
                in_flight_message = self.in_flight[e["Id"]]
                sec = int(e["VisibilityTimeout"])
                visible_at = now + datetime.timedelta(seconds=sec)
                updated_message = attr.evolve(in_flight_message, visible_at=visible_at)
                self.in_flight[e["Id"]] = updated_message
            else:
                not_found_entries.append(e)

        return {
            "Successful": [{"Id": e["Id"]} for e in found_entries],
            "Failed": [{"Id": e["Id"]} for e in not_found_entries],
        }

    def delete(self):
        return self.aws.delete_queue(self.url)

    def __dict__(self):
        return {"QueueUrl": self.url}

    def __getitem__(self, item):
        return self.__dict__()[item]


@attr.s(frozen=True)
class MemoryMessage:
    """
    A mock class to mimic the AWS message

    Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
         services/sqs.html#SQS.Message
    """

    queue_impl: MemoryQueue = attr.ib(repr=False)

    # The message's contents (not URL-encoded).
    body: bytes = attr.ib()

    # Each message attribute consists of a Name, Type, and Value.
    message_attributes: Dict[str, Dict[str, str]] = attr.ib(factory=dict)

    # A map of the attributes requested in `` ReceiveMessage `` to their
    # respective values.
    attributes: Dict[str, Any] = attr.ib(factory=dict)

    # Internal attribute which contains the execution time.
    visible_at: datetime.datetime = attr.ib(factory=datetime.datetime.utcnow)

    # A unique identifier for the message
    message_id: str = attr.ib(factory=lambda: uuid.uuid4().hex)

    # The Message's receipt_handle identifier
    receipt_handle: str = attr.ib(factory=lambda: uuid.uuid4().hex)

    @classmethod
    def from_kwargs(cls, queue_impl, kwargs):
        """
        Make a message from kwargs, as provided by queue.send_message():

        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
             services/sqs.html#SQS.Queue.send_message
        """
        # required attributes
        body = kwargs["MessageBody"]
        message_atttributes = kwargs.get("MessageAttributes")

        # optional attributes
        attributes = {"ApproximateReceiveCount": 1}
        if "MessageDeduplicationId" in kwargs:
            attributes["MessageDeduplicationId"] = kwargs["MessageDeduplicationId"]

        if "MessageGroupId" in kwargs:
            attributes["MessageGroupId"] = kwargs["MessageGroupId"]

        visible_at = datetime.datetime.utcnow()
        if "DelaySeconds" in kwargs:
            delay_seconds_int = int(kwargs["DelaySeconds"])
            visible_at += datetime.timedelta(seconds=delay_seconds_int)

        return MemoryMessage(
            queue_impl, body, message_atttributes, attributes, visible_at
        )

    def change_visibility(self, VisibilityTimeout="0", **kwargs):
        if self.message_id in self.queue_impl.in_flight:
            now = datetime.datetime.utcnow()
            sec = int(VisibilityTimeout)
            visible_at = now + datetime.timedelta(seconds=sec)
            updated_message = attr.evolve(self, visible_at=visible_at)
            self.queue_impl.in_flight[self.message_id] = updated_message
        else:
            logger.warning("Tried to change visibility of message not in flight")



import datetime

now = datetime.datetime.utcnow


class NeverShutdown:
    """Never shutdown the worker"""

    def update_state(self, batch_processing_result) -> None:
        pass

    def need_shutdown(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "NeverShutdown()"


class IdleShutdown:
    """Shutdown worker if it's idle for certain time (set in seconds)"""

    def __init__(self, idle_seconds: int) -> None:
        self.idle_seconds = idle_seconds
        self._idle_delta = datetime.timedelta(seconds=idle_seconds)
        self._is_idle = False
        self._last_seen = now()

    def update_state(self, batch_processing_result) -> None:
        """Update internal state of the shutdown policy"""
        if batch_processing_result.total_count() == 0:
            self._is_idle = True
        else:
            self._is_idle = False
            self._last_seen = now()

    def need_shutdown(self) -> bool:
        if not self._is_idle:
            return False
        return now() - self._last_seen >= self._idle_delta

    def __repr__(self) -> str:
        return f"IdleShutdown({self.idle_seconds})"


class MaxTasksShutdown:
    """
    Shutdown worker if it executed more than max_tasks in total (both
    successfully or with error)
    """

    def __init__(self, max_tasks: int) -> None:
        self.max_tasks = max_tasks
        self._tasks = 0

    def update_state(self, batch_processing_result) -> None:
        self._tasks += batch_processing_result.total_count()

    def need_shutdown(self) -> bool:
        return self._tasks >= self.max_tasks

    def __repr__(self) -> str:
        return f"MaxTasksShutdown({self.max_tasks})"


class OrShutdown:
    """Return True if any of conditions of policies is met"""

    def __init__(self, *policies) -> None:
        self.policies = policies

    def update_state(self, batch_processing_result) -> None:
        for p in self.policies:
            p.update_state(batch_processing_result)

    def need_shutdown(self) -> bool:
        return any(p.need_shutdown() for p in self.policies)

    def __repr__(self) -> str:
        repr_policies = ", ".join([repr(p) for p in self.policies])
        return f"OrShutdown({repr_policies})"


class AndShutdown:
    """Return True if all of conditions of policies are met"""

    def __init__(self, *policies) -> None:
        self.policies = policies

    def update_state(self, batch_processing_result) -> None:
        for p in self.policies:
            p.update_state(batch_processing_result)

    def need_shutdown(self) -> bool:
        return all(p.need_shutdown() for p in self.policies)

    def __repr__(self) -> str:
        repr_policies = ", ".join([repr(p) for p in self.policies])
        return f"AndShutdown({repr_policies})"


NEVER_SHUTDOWN = NeverShutdown()
ASAP_SHUTDOWN = IdleShutdown(0)



import base64
import json
import pickle
import zlib
from typing import Any, ClassVar, Dict, Protocol, Type

DEFAULT_CONTENT_TYPE = "pickle_compat"

# Highest version of the protocol, understood by Python2.
PICKLE_PY2_COMPAT_PROTO = 2


class Codec(Protocol):
    @classmethod
    def serialize(cls, message: Any) -> str:
        ...

    @classmethod
    def deserialize(cls, serialized: str) -> Any:
        ...


class JSONCodec:
    @classmethod
    def serialize(cls, message: Any) -> str:
        return json.dumps(message)

    @classmethod
    def deserialize(cls, serialized: str) -> Any:
        return json.loads(serialized)


class PickleCodec:
    protocol: ClassVar[int] = pickle.DEFAULT_PROTOCOL

    @classmethod
    def serialize(cls, message: Any) -> str:
        binary_data = pickle.dumps(message, protocol=cls.protocol)
        compressed_data = zlib.compress(binary_data)
        return base64.urlsafe_b64encode(compressed_data).decode("latin1")

    @classmethod
    def deserialize(cls, serialized: str) -> Any:
        compressed_data = base64.urlsafe_b64decode(serialized.encode("latin1"))
        binary_data = zlib.decompress(compressed_data)
        return pickle.loads(binary_data)


class PickleCompatCodec(PickleCodec):
    protocol = PICKLE_PY2_COMPAT_PROTO


def get_codec(content_type: str) -> Type[Codec]:
    return CONTENT_TYPES_CODECS[content_type]


CONTENT_TYPES_CODECS: Dict[str, Type[Codec]] = {
    "json": JSONCodec,
    "pickle": PickleCodec,
    "pickle_compat": PickleCompatCodec,
}
